/*  swefns.h
    Bryan Guillaume & Tom Nichols
    Copyright (C) 2019 University of Oxford  */

#ifndef ____swefns__
#define ____swefns__


#include <armadillo>
#include "newimage/newimageall.h"
#include "armawrap/newmat.h"
#include "cprob/libprob.h"
#include "sweopts.h"


namespace SWE {

  ////////////////////////////////////////
  // Functions related to matrix algebra
  ////////////////////////////////////////

  // compute the duplication matrix converting vech(A) into vec(A)
  arma::mat duplicationMatrix(arma::uword n);

  // compute the elimination indices which transform vec(A) into vech(A)
  arma::uvec eliminationIndices(arma::uword n);

  ////////////////////////////////////////
  // check from vech(A) if A is positive semidefinite (note that the function can cbind several vech(A) into a matrix) using a row reduction approach and, if not, make vechA positive semidefinite by zeroing the negative eigenvalues
  ////////////////////////////////////////

  void checkAndMakePSD(arma::mat& vechA);

  ////////////////////////////////////////
  // Temporary function allowing to combine arma::mat objects with NEWMAT::Matrix objects
  ////////////////////////////////////////

  // convert a NEWMAT::Matrix into an arma::mat
  arma::mat matrix2Mat(NEWMAT::Matrix& newmatMatrix);

  // convert an arma::mat into a NEWMAT::Matrix
  NEWMAT::Matrix mat2Matrix(arma::mat& armaMat);

  //	// convert a NEWMAT::Matrix into an arma::vec
  //	vec matrix2vec(NEWMAT::Matrix& newmatMatrix);
  //
  //	// convert a NEWMAT::Matrix into an arma::uvec
  //	uvec matrix2uvec(NEWMAT::Matrix& newmatMatrix);

  arma::vec columnVector2vec(NEWMAT::ColumnVector& newmatColumnVector);

  arma::uvec columnVector2uvec(NEWMAT::ColumnVector& newmatColumnVector);


  ////////////////////////////////////////
  // Template for the intersection of 2 arma vectors
  ////////////////////////////////////////
  template< typename T, template <typename> class Type >
  Type<T> vintersection(Type<T> a, Type<T> b)
  {
    std::vector<T> o;
    std::set_intersection(a.begin(), a.end(), b.begin(), b.end(), std::back_inserter(o)) ;
    std::reverse(o.begin(), o.end()) ;

    Type<T> result = arma::conv_to< Type<T> >::from(o);
    return result;
  }

  ////////////////////////////////////////
  // Function to read vest format and convert the info into a mat object
  ////////////////////////////////////////

  // inspired from MISCMATHS::read_vest (needed to be changed to allow armadillo matric class)
  arma::mat read_vest_swe(std::string p_fname);

  void T2ZP(arma::mat& t, arma::mat& z, arma::mat& pValues, const arma::mat& dof, const bool logP);

  void F2XP(arma::mat& f, arma::mat& x, arma::mat& pValues, double dof1, const arma::mat& dof2, const bool logP);
  void F2XZP(arma::mat& f, arma::mat& x, arma::mat& z, arma::mat& pValues, double dof1, const arma::mat& dof2, const bool logP);

  ////////////////////////////////////////
  // Class related voxelwise designs
  ////////////////////////////////////////

  // VoxelwiseDesign is a class similar to RANDOMISE::VoxelwiseDesign, but using armadillo instead of newmat
  class VoxelwiseDesign
  {
  public:
    bool isSet;
    arma::field<arma::mat> EV;
    arma::uvec  location;
    void setup(const std::vector<int>& voxelwise_ev_numbers, const std::vector<std::string>& voxelwise_ev_filenames, const NEWIMAGE::volume<float>& mask, const int maximumLocation, const bool isVerbose);
    VoxelwiseDesign() { isSet=false; }
    arma::mat adjustDesign(const arma::mat& originalDesign, const int voxelNo);
  private:
  };

  ////////////////////////////////////////
  // Class containing info about the SwE Model
  ////////////////////////////////////////

  class SweModel
  {
  public:
    bool modified; // modifed SwE or classic SwE
    bool wb; // non parametric Wild Bootstrap or parametric inference
    bool voxelWiseDesign; // indicate if there are voxelwise covariates or not
    bool voxelwiseOutput; // output voxelwise corrected p-values
    bool doFOnly; // indicate if only F-contrasts are considered
    bool doTOnly; // indicate if only F-contrasts are considered
    bool verbose;
    bool logP; // if false, the p-value images will be 1-p; if true, the p-value images will be -log10(p)
    bool outputRaw; // output t & f statistic images
    bool outputEquivalent; // output z & x statitic images
    bool outputDof; // output effective number of degrees of freedom images
    bool outputUncorr; // output voxelwise uncorrected p-values
    bool outputGlm; // output pe, cope & varcope images
    bool outputTextNull; // output a text file with max. stats
    bool doTfce; // If true, will run a TFCE inference

    std::string out_fileroot;
    arma::uword nVox; // total number of in-mask voxels
    arma::uword nData; // total number of input images
    arma::mat designMatrix;
    VoxelwiseDesign voxelWiseEV;
    arma::mat pinvDesignMatrix;
    NEWIMAGE::volume<float> mask;
    arma::mat tc;
    arma::mat fc;
    arma::uword nContrasts;
    arma::uword startFC;
    arma::field<arma::mat> fullFContrast;
    arma::uvec sizeFcontrast;
    arma::uvec nCovCFSweCF;
    arma::field<arma::rowvec> weightT; // link the contrasted SwE to the subject (or group) covariance matrices for t-contrasts
    arma::field<arma::mat> weightF; // link the contrasted SwE to the subject (or group) covariance matrices for F-contrasts
    arma::uvec subject;
    arma::uvec uSubject;
    arma::uvec visit;
    arma::uvec group;
    arma::uvec uGroup;
    arma::field<arma::uvec> uVisit; // contains the unique visit category labels for each group (for modified SwE)
    arma::uvec n_i; // number of visit per subject
    arma::uvec nCovVis_i; // number of unique variances/covariances per subject
    arma::uvec n_g; // number of visit per groups
    arma::uvec nCovVis_g; // number of unique variances/covariances per group (for modified SwE)
    arma::field<arma::uvec> subjectIndex;
    arma::field<arma::umat> subjectCovIndex;
    arma::vec subjectDof;
    arma::uvec iGrDof; // indicates which separate design matrix each row (or observation) belongs to
    arma::uvec indDiag;
    arma::uvec indOffDiag;
    arma::umat indCorrDiag;
    arma::field<arma::uvec> indDiagRes;
    arma::field<arma::umat> indOffDiagRes;
    arma::field<arma::mat> dofMat;
    arma::uword nBootstrap;
    float clusterThresholdT;
    float clusterThresholdF;
    float clusterMassThresholdT;
    float clusterMassThresholdF;
    arma::mat resamplingMult;
    float tfceHeight;
    float tfceExtent;
    bool tfceDeltaOverride;
    arma::vec tfceDelta;
    arma::uword tfceConnectivity;

    NEWIMAGE::volume<float> nonConstantMask(NEWIMAGE::volume4D<float>& data, const bool allOnes);
    void computeSubjectDof();
    void computeWeightClassic();
    void computeWeightModified();
    void setup(const sweopts& opts);
    void checkModel(const arma::uword numberOfDataPoints); // check if the model looks alright
    void loadData(arma::mat& data, const sweopts& opts);
    void loadVoxelWiseDesign(const sweopts& opts);
    void adjustDofMat();
    void adjustDesign(const arma::uword iVox); // adjust the design for the voxel iVox (only useful for voxelwise designs)
    arma::mat computePe(const arma::mat& data); // compute the parameter estimates
    arma::mat computeAdjustedResiduals(const arma::mat& data, const arma::mat& beta);
    arma::mat computeAdjustedRestrictedResiduals (const arma::mat& data, const arma::mat& beta, const arma::mat& contrast);
    arma::mat computeGroupCovarianceMatrices();
    void saveImage(arma::mat& dataToSave, std::string filename); // cannot compile with const argument...To check later
    void saveImage(arma::subview_row<double>& dataToSave, std::string filename);
    void computeVechCovVechV(arma::mat& VechCovVechV, const arma::mat& covVis_g, const arma::uword g);
    void computeClusterStatsT(arma::mat& score, arma::uvec& clusterLabels, arma::uvec& clusterExtentSizes);
    void computeClusterStatsT(arma::mat& score, arma::uvec& clusterLabels, arma::vec& clusterMassSizes);
    void computeClusterStatsT(arma::mat& pValue, arma::mat& score, arma::uvec& clusterLabels, arma::vec& clusterMassSizes);
    void computeClusterStatsF(arma::mat& score, arma::uvec& clusterLabels, arma::uvec& clusterExtentSizes);
    void computeClusterStatsF(arma::mat& score, arma::uvec& clusterLabels, arma::vec& clusterMassSizes);
    void computeClusterStatsF(arma::mat& pValue, arma::mat& score, arma::uvec& clusterLabels, arma::vec& clusterMassSizes);
    void tfce(arma::mat& zScore, arma::mat& tfceScore, const arma::uword indexContrast);
    void generateResamplingMatrix(); // generate resampling matrix for the WB
    void printMaxStats(const arma::vec& maxStat, const std::string typeLabel, const std::string statLabel); // print maximum statistics
    void saveClusterWisePValues(const arma::mat& oneMinusFwePClusterMass, const arma::uvec& clusterMassLabels, const std::string typeLabel, const std::string statLabel);
    void computeSweT(arma::mat& swe, const arma::mat& residuals, const arma::uword iContrast); 	// compute the modified SwE for t-contrasts
    void computeSweT(arma::mat& swe, arma::mat& dof, const arma::mat& residuals, const arma::uword iContrast); 	// compute the modified SwE for t-contrasts
    void computeSweF(arma::mat& swe, const arma::mat& residuals, const arma::uword iContrast); 	// compute the modified SwE for f-contrasts
    void computeSweF(arma::mat& swe, arma::mat& dof, const arma::mat& residuals, const arma::uword iContrast); 	// compute the modified SwE for f-contrasts
    void computeT(const arma::mat& data, arma::mat& tstat, arma::mat& cope, arma::mat& varcope, const arma::uword iContrast); 	// compute t-contrasts
    void computeT(const arma::mat& data, arma::mat& tstat, arma::mat& cope, arma::mat& varcope, arma::mat& dof, const arma::uword iContrast); // compute t-contrasts
    void computeF(const arma::mat& data, arma::mat& fstat, arma::mat& cope, arma::mat& varcope, const arma::uword iContrast); // compute f-contrasts (no effective degrees of freedom adjustments)
    void computeF(const arma::mat& data, arma::mat& fstat, arma::mat& cope, arma::mat& varcope, arma::mat& dof, const arma::uword iContrast); // compute f-contrasts
    void computeSweAll(arma::field<arma::mat>& swe, const arma::mat& residuals); // compute the SwE for all contrasts
    void computeSweAll(arma::field<arma::mat>& swe, arma::mat& dof, const arma::mat& residuals); // compute the SwE and dof for all contrasts
    void computeSweAll(arma::field<arma::mat>& swe, const arma::mat& residuals, const arma::uword iVox); // compute the SwE for all contrasts, but only voxel iVox
    void computeSweAll(arma::field<arma::mat>& swe, arma::mat& dof, const arma::mat& residuals, const arma::uword iVox); // compute the SwE and dof for all contrasts, but only voxel iVox
    void computeStatsClassicSwe(const arma::mat& data); // estimate the model & stats for the classic SwE (parametric inferences)
    void computeStatsModifiedSwe(const arma::mat& data); // estimate the model & stats for the modified SwE (parametric inferences)
    void computeStatsWb(const arma::mat& data); // estimate the model & statsf using the WB procedure

  private:
  };

} // end of namespace SWE

#endif /* defined(____swedesign__) */
